import signal
from time import sleep

from .args import config
from .logging import logger
from .utils import clean_existing_files, \
    convert_existing_files, get_all_files, convert_existing_file, get_exclude_dirs
from .watchdog import schedule_observer, get_observer
from multiprocessing.dummy import Pool as ThreadPool

observer = get_observer()


def finish(signum, frame):
    logger.info('Exiting application...')
    try:
        observer.stop()
        observer.join()
    except:
        logger.info('Error stopping observer.')
    exit(0)


signal.signal(signal.SIGTERM, finish)
signal.signal(signal.SIGINT, finish)

if config.dry_run:
    logger.info('Dry run enabled.  Will not perform any file operations but will still logger.info output of what would be happening.')

for p in get_exclude_dirs():
    logger.info(f'Excluding directory: {p}')


# Only start watcher if no subcommand is specified or watch command is specified
if config.subcommand is None or config.subcommand == 'watch':
    logger.info('Starting watcher')
    observer = schedule_observer(observer)
    observer.start()

    # New thread pool to process multiple existing files simultaneously
    pool = ThreadPool(config.workers)

    # Run new thread pool from all existing files
    logger.info('Running conversion on existing files.')
    results = pool.map(convert_existing_file, get_all_files())
    logger.info('Finished processing existing files.')
    # Run indefinitely so watchdog can do its thing
    while True:
        sleep(1)

elif config.subcommand == 'clean':
    logger.info('Cleaning existing files')
    clean_existing_files(get_all_files())

elif config.subcommand == 'convert':
    logger.info('Converting existing files')
    convert_existing_files(get_all_files())



import os
from pathlib import Path
from time import sleep

from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler
from watchdog.observers.polling import PollingObserver

from .MediaFile import MediaFile
from .args import config, IS_DOCKER
from .logging import logger
from .utils import get_exclude_dirs


def on_created(event):
    logger.info(f"CREATED: {event.src_path}")
    size = -1
    while size != os.path.getsize(event.src_path):
        size = os.path.getsize(event.src_path)
        sleep(1)
    media = MediaFile(Path(event.src_path))
    media.process_file()


def get_observer():
    observer = Observer()
    if config.windows and IS_DOCKER:
        logger.info('WARNING: Using less performant observer because host is Windows.')
        observer = PollingObserver()
    return observer


def schedule_observer(observer):
    fs_handle = PatternMatchingEventHandler(
        patterns=['*'],
        ignore_patterns=get_exclude_dirs(),
        ignore_directories=False,
        case_sensitive=True
    )
    fs_handle.on_created = on_created

    for p in config.watch_dirs:
        path = Path(p)
        if path.exists():
            logger.info(f'Watching directory {path.resolve()}')
            observer.schedule(fs_handle, f'{path.resolve()}', recursive=True)
        else:
            logger.info(f'Skipping watch directory `{path.resolve()}` because it does not exist')
    return observer




import argparse
import os
from pathlib import Path
from environs import Env

env = Env()
IS_DOCKER = env.bool('IS_DOCKER', False)
IS_WINDOWS = env.bool('IS_WINDOWS', False)
_parser = argparse.ArgumentParser(description='Convert media to smaller web formats.', prog='webwatcher')


class AppendOrOverwrite(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None)
        items = argparse._copy_items(items)
        items.append(values)

        # Remove `/watch` if user specified their own watch directory(s)
        if len(items) > 1 and '/watch' in items:
            items.remove('/watch')
        setattr(namespace, self.dest, items)


def quality(arg):
    """ Type function for argparse - a float within some predefined bounds """
    try:
        f = int(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be an integer")
    if f < 1 or f > 100:
        raise argparse.ArgumentTypeError("Argument must be < " + str(100) + "and > " + str(1))
    return str(f)


def file_extension(arg):
    """ Type function for argparse - a float within some predefined bounds """
    if str(arg).startswith('.'):
        return str(arg)
    raise argparse.ArgumentTypeError('Argument must be a file extension starting with \'.\'')


_WATCH_DIRS = env.list('WATCH_DIRS', list())
if len(_WATCH_DIRS) == 0:
    _WATCH_DIRS.append('/watch')
WATCH_DIRS = [Path(f) for f in _WATCH_DIRS]

_EXCLUDE_DIRS = env.list('EXCLUDE_DIRS', list())
EXCLUDE_DIRS = [Path(f) for f in _EXCLUDE_DIRS]

DRY_RUN = env.bool('DRY_RUN', False)
COPY_SOURCE = env.bool('COPY_SOURCE', True)
SOURCE_DIR = env.path('SOURCE_DIR', '/source')
BASE_DIR = env.path('BASE_DIR', '/watch')
WORKERS = env.int('WORKERS', 8)
DELETE_MEDIA = env.bool('DELETE_MEDIA', True)

# Audio
WATCH_AUDIO = env.bool('WATCH_AUDIO', True)
DELETE_AUDIO = env.bool('DELETE_AUDIO', True)
_AUDIO_CONVERT_FORMATS = ['.mp3', '.aac', '.flac', '.wav', '.wma', '.aac', '.m4a', '.ogg']

# Image conversion settings
WATCH_IMAGES = env.bool('WATCH_IMAGES', True)
DELETE_IMAGES = env.bool('DELETE_IMAGES', True)
_IMAGE_CONVERT_FORMATS = ['.png', '.jpg', '.bmp', '.jpeg']
WEBP_COMMAND = env.str('WEBP_COMMAND', 'magick' if os.name == 'nt' else 'convert')
WEBP_LOSSLESS = env.bool('WEBP_LOSSLESS', False)
WEBP_QUALITY = env.str('WEBP_QUALITY', '100' if WEBP_LOSSLESS else '60')


# Define subcommands
subparsers = _parser.add_subparsers(help='other functions', title='subcommands', dest='subcommand', metavar='{command}')
# Watch command
p_manage = subparsers.add_parser('watch', help='Watches one or more directories for new files to convert.', )
# Clean command
p_manage = subparsers.add_parser('clean', help='Cleans up old files that may or may not have been converted yet.')
# Convert command
p_manage = subparsers.add_parser('convert', help='Runs a one time conversion of all matching files in the specified directories.')


# General args
_parser.add_argument('-p', '--path', type=Path, dest='watch_dirs', action=AppendOrOverwrite, metavar='directory', help='a patch to watch for files', default=_WATCH_DIRS)
_parser.add_argument('-d', '--dry-run', action='store_true', help='Do not process any files but show output.', default=DRY_RUN)
_parser.add_argument('--no-copy-source', action='store_false', dest='copy_source', help='Do not copy source files to source folder.', default=COPY_SOURCE)
_parser.add_argument('--windows', action='store_true', help='Is running on Windows host. (only required when running from Docker)', default=IS_WINDOWS)
_parser.add_argument('--delete-media', action='store_true', help='Delete media files when done with them.', default=DELETE_MEDIA)
_parser.add_argument('-s', '--source-dir', type=Path, action='store', metavar='directory', help='directory to place source files (should be outside all watch directories)', default=SOURCE_DIR)
_parser.add_argument('-b', '--base-dir', type=Path, action='store', metavar='directory', help='sets base watch directory (debug only)', default=BASE_DIR)
_parser.add_argument('-e', '--exclude', type=Path, dest='exclude_dirs', action=AppendOrOverwrite, metavar='directory', help='Paths to exclude from processing', default=EXCLUDE_DIRS)
_parser.add_argument('-w', '--workers', type=int, action='store', metavar='int', help='max number of worker threads to use for simultaneous processing.', default=WORKERS)
_parser.add_argument('--all', action='store_true', help='Deletes ALL files, even if it isn\'t a media file (excluding webm/webp files)')
_parser.add_argument('--force', action='store_true', help='Deletes matching media files, even if the converted webm/webp is not found.')

# Audio args
a_group = _parser.add_argument_group('Audio')
a_group.add_argument('--no-watch-audio', action='store_false', dest='watch_audio', help='Do not watch for audio files', default=WATCH_AUDIO)
a_group.add_argument('--no-delete-audio', action='store_true', dest='keep_audio', help='Do not delete audio (used with --delete-media)', default=DELETE_AUDIO)
a_group.add_argument('--audio-format', type=file_extension, action='append', dest='audio_convert_formats', metavar='extension', help='Extra audio formats to watch for', default=_AUDIO_CONVERT_FORMATS)


# Image args
i_group = _parser.add_argument_group('Images')
i_group.add_argument('--no-watch-images', action='store_false', help='Do not watch for image files', default=WATCH_IMAGES)
i_group.add_argument('--no-delete-images', action='store_true', dest='keep_images', help='Do not delete images (used with --delete-media)', default=DELETE_IMAGES)
i_group.add_argument('--image-format', type=file_extension, action='append', dest='image_convert_formats',metavar='extension', help='Extra image formats to watch for', default=_IMAGE_CONVERT_FORMATS)
i_group.add_argument('--webp-command', type=str, nargs='?', metavar='executable_path', help='command to run for ImageMagick', default=WEBP_COMMAND)
i_group.add_argument('--webp-quality', type=quality, action='store', metavar='percent', help='conversion quality for libwebp: 1 (worst) to 100 (lossless)', default=WEBP_QUALITY)
i_group.add_argument('--webp-lossless', action='store_true', help='Use lossless conversion when converting to webp', default=WEBP_LOSSLESS)

config = _parser.parse_args()








import os
import shutil
import signal
import subprocess
import sys
from pathlib import Path, WindowsPath
import ffmpeg

from webwatcher.args import config, IS_DOCKER
from webwatcher.logging import logger


class MediaFile:

    def __init__(self, path: Path, parent: Path = None):
        self.path = path
        self.parent = parent if parent else path
        self.is_media = False
        self.converted = False
        self.delete = config.delete_media
        self.keep = not self.delete
        self.dest = path
        self.failed = False

        # Set parent based on watch_dir
        for i, p in enumerate(self.path.parents):
            if p in config.watch_dirs:
                self.parent = p
                break

        # If only watch directory is /watch
        if len(config.watch_dirs) == 1 and config.watch_dirs[0] == '/watch':
            # Make source relative to watch directory
            self.source = config.source_dir / self.path.relative_to(self.parent.resolve())
        else:
            # Make source location relative to drive (to avoid duplicate paths)
            self.source = config.source_dir / self.path.relative_to(self.path.parents[-1].resolve())


        if self.path.suffix in config.audio_convert_formats:
            self.dest = self.dest.with_suffix('.webm')
            self.is_media = True
            self.is_audio = True
            self.keep = not config.keep_audio
        elif self.path.suffix in config.image_convert_formats:
            self.dest = self.dest.with_suffix('.webp')
            self.is_media = True
            self.is_image = True
            self.keep = not config.keep_images

        if not IS_DOCKER:
            self.source = config.source_dir / self.path.relative_to(self.path.anchor)

    def copy_to_source(self, override=False, move=False):
        """
        Copies the original file to the source folder, relative to the watch directory.
        :param override: boolean
        :param move: Move file instead of copying
        """
        if config.copy_source and not override:
            # Create source destination if not exists
            if not config.dry_run:
                self.source.parent.mkdir(exist_ok=True, parents=True)

            # Only copy if file doesn't exist already
            if not self.source.exists():
                logger.info(f'COPY:\t{self.source}')
                # Copy source to new folder
                if not config.dry_run:
                    if move:
                        shutil.move(self.path, self.source)
                    else:
                        shutil.copy(self.path, self.source)

    def convert(self):
        """
        Converts the input file based on its media type (audio or image)
        :return: success
        """
        # Exit if not media file
        if not self.is_media:
            return True

        if self.dest.suffix == '.webm':
            return self.convert_audio()
        elif self.dest.suffix == '.webp':
            return self.convert_image()

    def delete_file(self):
        # If the file was a supported extension
        logger.info(f'DELETE:\t{self.path.resolve()}')
        if not config.dry_run:
            self.path.unlink(missing_ok=True)

    def clean(self, copy=False):
        """
        Delete the source file it not in dry run mode
        """
        if self.is_media:
            if self.dest.exists() or config.force:
                if copy:
                    self.copy_to_source(move=True)
                self.delete_file()
        elif config.all:
            if self.path.suffix not in ['.webm', '.webp']:
                if copy:
                    self.copy_to_source(move=True)
                self.delete_file()

    def convert_audio(self, overwrite=False):
        """
        Process the audio file with ffmpeg
        :param overwrite: bool
        :return: success<bool>
        """
        if not self.dest.exists() or overwrite:
            short_args = [
                'ffmpeg', '-y',
                '-i', f'"{self.path.relative_to(self.parent).resolve()}"',
                '-map', '0:a',
                '-c:a', 'libopus',
                '-quality', '100',
                f'"{self.dest.relative_to(self.parent).resolve()}"'
            ]
            logger.info(' '.join(['AUDIO: ', *short_args]))
            if not config.dry_run:
                try:
                    (ffmpeg
                        .input(f'{self.path.resolve()}')
                        .output(f'{self.dest.resolve()}', map='0:a', format='webm',acodec='libopus', quality=100)
                        .run(capture_stdout=True, capture_stderr=True)
                    )
                    self.converted = True
                    return True
                except ffmpeg.Error as e:
                    logger.info(f'ffmpeg encountered an error converting audio file {self.path.resolve()}')
                    self.failed = True
                    return False
            return True

    def convert_image(self, overwrite=False):
        """
        Process the image file with ImageMagick
        :param overwrite: bool
        :return: success<bool>
        """
        code = 0
        # Only run conversion if destination file does not exist.
        if not self.dest.exists() or overwrite:
            args = [
                config.webp_command,
                f'{self.path.resolve()}',
                '-quality', config.webp_quality,
                f'{self.dest.resolve()}'
            ]
            short_args = [
                config.webp_command,
                f'"{self.path.relative_to(self.parent).resolve()}"',
                '-quality', config.webp_quality,
                f'"{self.dest.relative_to(self.parent).resolve()}"'
            ]
            logger.info(' '.join(['IMAGE: ', *short_args]))
            if not config.dry_run:
                p = subprocess.run(args)
                code = p.returncode
            if code == 0:
                self.converted = True
                return True
            else:
                self.failed = True
                return False

    def process_file(self):
        """
        Converts the file, followed by a copy and clean
        """
        self.convert()
        if not self.failed:
            self.clean(copy=True)




import glob
from pathlib import Path

from webwatcher.MediaFile import MediaFile
from webwatcher.args import config


def get_exclude_dirs():
    exclude = list()
    for path in config.exclude_dirs:
        if path.is_absolute():
            exclude.append(path)
        else:
            # Return list of dirs with watch_dir appended
            exclude.extend([watch / path for watch in config.watch_dirs])
    return exclude


def get_all_files():
    # Get all files from all watched dirs
    all_files = []
    for d in config.watch_dirs:
        watch_dir = Path(d)
        files = [Path(f) for f in glob.glob(f'{watch_dir.resolve()}/**/*', recursive=True) if Path(f).is_file()]
        for f in files:
            all_files.append((f, watch_dir))
    return all_files


def convert_existing_files(file_pairs):
    for path, parent in file_pairs:
        media = MediaFile(path, parent=parent)
        media.process_file()


def convert_existing_file(file_pair):
    path, parent = file_pair
    exclude = any(path.match(p) for p in ['test/*'])
    if not exclude:
        media = MediaFile(path, parent=parent)
        media.process_file()


def clean_existing_files(file_pairs):
    for path, parent in file_pairs:
        media = MediaFile(path, parent=parent)
        media.clean(copy=True)




from distutils.core import setup


setup(
    name='webwatcher',
    packages=['webwatcher'],
    version='1.2.18',
    license='MIT',
    description='Convert media files to a modern smaller equivalent.',
    long_description='For more information on how to use, visit the repo at https://gitlab.com/cclloyd1/webwatcher',
    author='cclloyd',
    author_email='cclloyd@cclloyd.com',
    url='https://gitlab.com/cclloyd1',
    keywords=['media', 'webp', 'webm', 'convert', 'compress'],
    install_requires=[
        'environs',
        'watchdog',
        'ffmpeg-python',
    ],
    python_requires='>=3.8',
    classifiers=[
        # "3 - Alpha"    "4 - Beta"    "5 - Production/Stable"
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Topic :: Multimedia :: Graphics :: Graphics Conversion',
        'Topic :: Multimedia :: Sound/Audio :: Conversion',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3 :: Only',
    ],
    setup_requires=['wheel'],
)



import logging

# Filter to pad levelname with spaces on right without padding inside brackets
def fmt_filter(record):
    record.levelname = '[%s]' % record.levelname
    record.funcName = '[%s]' % record.funcName
    return True


logging.basicConfig(
    format='WebWatcher | %(asctime)s | %(levelname)-10s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S',
)
logger = logging.getLogger('webwatcher')
logger.addFilter(fmt_filter)
